import numpy as np
import time
import functools
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d
from pvtrace import *
#todo: histogram of photon wavelengths at end
#todo: proof absorbance / absorption coefficient relationship:
#https://www.researchgate.net/post/How_can_I_calculate_the_
#Absorption_coefficient_from_Absorbance2



#################################################################
#			Input Parameteres			#
#################################################################
# DNK-B001 from Creative Diagnostics				#
conc=0.001			# QD concentration [nm^-3]	#
diam=2.8			# QD diameter [nm]		#
Qyield=0.6			# quantum yield			#
ref_ind=2.64			# QD refractive index		#
#	^ http://www.electrochemsci.org/papers/vol7/7098458.pdf	#
wg_l=0.5e2			# waveguide length [nm]		#
wg_w=0.5e2			# waveguide width [nm]		#
wg_h=0.5e2			# waveguide height [nm]		#
wg_ref=1.49			# waveguide refractive index	#
coat_l=10.			# coating thickness [nm]	#
coat_ref=1.3			# coating refractive index	#
photon_num=100000		# number of photons emitted	#
#################################################################

print('###########################')
print('‹‹‹ NOW READING IN DATA ›››')
print('###########################')

#################################################################
#			Read In Emission			#
#################################################################
# Read in emission spectra data from Zhang2017 script		#
# 	store force and emission spectra in emi_spec		#
f=open('emi_spec.txt','r')					#
print('Reading in emission spectra...')				#
emi_spec=[]							#
for line in f:							#
	if 'Force' in line:	# have we reached a force value?#
		print('Now at force: '+str(line.split()[1]))	#
		try:		# see if force variable ...	#
			force	# is defined			#
		except NameError:				#
				# if it is not, we are at start	#
				# of file, we define force value#
			force=float(line.split()[1])		#
		else:		# if it is, we save the spectra	#
			emi_spec.append([force,temp_emi_spec])	#
				# and define new force value	#
			force=float(line.split()[1])		#
		temp_emi_spec=[]# new force = new temp spectra	#
	else:			# read in spec to temp_emi_spec	#
		wave_len=float(line.split()[0])			#
		emission_val=float(line.split()[1])		#
		temp_emi_spec.append([wave_len,emission_val])	#
print('Finished reading in emission spectra')			#
#print(len(emi_spec))		# number of force steps		#
#print(len(emi_spec[0]))	# dimensions of force step 1	#
#print(emi_spec[0][0])		# force step 1 force		#
#print(emi_spec[0][1])		# force step 1 spectrum		#
#print(len(emi_spec[0][1][0]))	# wavelength and value of the	#
				# first wavelength in spectrum	#
#################################################################



#################################################################
#			Read In Absorption			#
#################################################################
# Read in and interpolate digitalised absorbance spectrum from	#
# https://pubs.acs.org/doi/full/10.1021/ja0782706		#
# "Quantum Dot Solar Cells. Tuning Photoresponse through Size 	#
# 	and Shape Control of CdSe−TiO2 Architecture"		#
# 	by Kongkanand, Tvrdy, Takechi, Kuno and Kamat		#
# 	Figure 1. (3.0 nm)					#
g=open('abs_spec.txt','r')					#
print('Reading in absorbance spectrum')				#
digi_wl=[]			# store digitalised wavelength	#
digi_abs=[]			# store digitalised absorbance	#
for line in g:							#
	digi_wl.append(float(line.split(", ")[0]))		#
	digi_abs.append(float(line.split(", ")[1]))		#
print('Finished reading in absorbance spectrum')		#
print('Extrapolating absorbance spectrum...')			#
f=interp1d(digi_wl,digi_abs,fill_value="extrapolate")		#
				# create wavelength list	#
wl=[emi_spec[0][1][i][0] for i in range(0,len(emi_spec[0][1]))]	#
absorbance=f(wl)		# create absorbance		#
print('Finished extrapolating absorbance spectrum')		#
absorption=[]			# create absorption		#
print('Converting absorbance into absorption')			#
for index in range(len(absorbance)):				#
	if absorbance[index]<0.:# make sure its positive	#
		absorption.append(0.)				#
	else:							#
		absorption.append(2.303*absorbance[index]/3.)	#
abs_spec=np.column_stack((wl,absorption))			#
print('Finished converting into absorption')			#
#print(abs_spec)						#
#print(len(abs_spec))		# number of wavelengths		#
#print(abs_spec[0])		# first wavelength and abs	#
#print(abs_spec[0][0])		# first wavelength		#
#plt.plot(wl, f(wl))						#
#plt.show()							#
#################################################################


print('############################')
print('‹‹‹ NOW CREATING OBJECTS ›››')
print('############################')

#################################################################
#			Setup Materials				#
#################################################################
# Create environments: world, waveguide coating, waveguide core #
# World taken as air						#
# Coating environment: WACKER SilGel 612 A/B			#
# Absorption = 0 (Datasheet states "crystal clear")		#
# Core environment: silicone LS1-3252 				#
# Absorption = 0 ('Optically clear')				#
				# create world			#
print('Creating world space')					#
world=Node(name='world (air)',geometry=Box(size=(2.*wg_l,2.*wg_w\
	,1.5*wg_h),material=Material(refractive_index=1.0)))	#
print('Finished creating world space')				#
print('Creating waveguide coating')				#
				# create coating		#
coat=Node(name='coating (SilGel)',geometry=Box(size=(wg_l+coat_l\
	,wg_w+coat_l,wg_h),material=Material(refractive_index= 	\
	1.4)),parent=world)					#
print('Finished creating waveguide coating')			#
coat.translate((coat_l/2.,0.,0.))				#
print('Creating waveguide core')				#
				# create core			#
core=Node(name='core (Si)',geometry=Box(size=(wg_l,wg_w,wg_h),	\
	material=Material(refractive_index=1.5)),parent=coat)	#
print('Finished creating waveguide core')			#
#################################################################


print('########################')
print('‹‹‹ NOW CREATING QDS ›››')
print('########################')


#################################################################
#			Set Up QD Mesh				#
#################################################################
# First the volume in which the center of these QDs		#
# 	can exist (vol of core minus the radius of QDs at edges)#
#	is found, and a 3D rectangular mesh is created, with	#
#	distance between each neighbouriing point being		#
#	concentration^-1/3					#
print('Setting up QD location mesh')				#
mesh_x=np.arange(-(wg_l-diam)/2.,(wg_l-diam)/2.,conc**(-1./3.))	#
mesh_y=np.arange(-(wg_w-diam)/2.,(wg_w-diam)/2.,conc**(-1./3.))	#
mesh_z=np.arange(-(wg_h-diam)/2.,(wg_h-diam)/2.,conc**(-1./3.))	#
print('Finished setting up QD location mesh')			#
#mesh=np.meshgrid(mesh_x,mesh_y,mesh_z)				#
#for i in mesh_x:						#
#	for j in mesh_y:					#
#		for k in mesh_z:				#
#			print(i,j,k)				#
# Function to populate core with QDs                            #
# A function is used because each QD can only take on one       #
#       emission spectrum.                                      #
# Therefore, at each force value, we call this function to      #
#       populate the core with QDs of different emission spectra#
#################################################################



#################################################################
#		Iterate Force Steps and QDs			#
#################################################################
#for force in range(len(emi_spec)):				#
#	d={}			# dictionary to store QDs	#
#	for i in mesh_x:					#
#		for j in mesh_y:				#
#			for k in mesh_z:			#
#			d['QD_{0}_{1}_{2}'.format(i,j,k)]=Node(	\
#	name='CdSe_[0]_{1}_{2}'.format(i,j,k),geometry=Sphere(	\
#	radius=diam/2.),material=Material(refractive_index=2.64,\
#	components=[Luminophore(coefficient=abs_spec,emission=	\
#	emi_spec[force][1],quantum_yield=Qyield)]),parent=core)	#
d={}                    	# dictionary to store QDs       #
qd_num=0			# label to number quantum dots	#
for i in mesh_x:						#
	for j in mesh_y:					#
		for k in mesh_z:				#
			print('Creating QD: '+str(qd_num+1))	#
			QD=Node(name='CdSe_{0}'.format(qd_num),	\
			geometry=Sphere(radius=diam/2.,material=\
			Material(refractive_index=2.64,		\
			components=[Luminophore(coefficient=	\
			abs_spec,emission=np.array(emi_spec	\
			[0][1]),quantum_yield=Qyield)])),parent\
			=core)					#
			print('Translating QD')			#
			QD.translate((i,j,k))			#
			qd_num=qd_num+1				#
print('All QDs created')					#
#################################################################


print('###########################')
print('‹‹‹ NOW CREATING LIGHT  ›››')
print('###########################')


#################################################################
#			Add source of photons			#
#################################################################
				# add lightcone of 30 radians	#
print('Creating light cone')					#
light=Node(name="Light (425nm)",parent=world,light=Light(	#
	wavelength=425,direction=functools.partial(cone,np.	\
	radians(30))))						#
light.translate((0,0,-wg_h/2.))					#
print('Finished creating light cone')				#
#################################################################



#################################################################
#		Use meshcat to render the scene			#
#################################################################
#viewer=MeshcatRenderer(open_browser=False,transparency=False,	\
#	opacity=0.5,wireframe=True)				#
scene=Scene(world)						#
out=open('ray_output.txt','w+')					#
#viewer.render(scene)						#
pos_x=[]							#
pos_y=[]							#
dir_x=[]							#
dir_y=[]							#
dir_z=[]							#
ray_l=[]							#
ray_num=1							#
for ray in scene.emit(photon_num):				#
	print('Tracing ray: '+str(ray_num))			#
	ray_num=ray_num+1					#
	history=photon_tracer.follow(scene,ray)			#
	path,events=zip(*history)				#
#	viewer.add_ray_path(path)				#
	if history[-2][0].position[2]==wg_h/2.:			#
		print('Ray reached end of waveguide')		#
#		print(history[-2][1].name)			#
		if history[-2][1].name=='TRANSMIT':		#
			print('Ray is transmitted')		#
			print('Writing to output file...')	#
			pos_x.append(float(history[-2][0].	\
			position[0]))				#
			pos_y.append(float(history[-2][0].	\
			position[1]))				#
			dir_x.append(float(history[-2][0].	\
			direction[0]))				#
			dir_y.append(float(history[-2][0].	\
			direction[1]))				#
			dir_z.append(float(history[-2][0].	\
			direction[2]))				#
			ray_l.append(float(history[-2][0].	\
			wavelength))				#
			out.write(str(history[-2][0].position[0]\
			)+'	'+str(history[-2][0].position[1]\
			)+'	'+str(history[-2][0].direction[0\
			])+'	'+str(history[-2][0].direction[1\
			])+'	'+str(history[-2][0].direction[2\
			])+'	'+str(history[-2][0].wavelength)\
			+'\n')					#
	else:							#
		print('Ray did not reach end of waveguide')	#
out.write('imdone')
#print(pos_x,pos_y,dir_x,dir_y,dir_z,ray_l)			#
#while True:			# Keep script alive until Ctrl-C#
#	try:							#
#		time.sleep(0.1)					#
#	except KeyboardInterrupt:				#
#		break						#
#################################################################


#plt.subplot(2,3,1)
#plt.xlabel('X coordinate')
#plt.hist(pos_x,bins=20)
#plt.subplot(2,3,2)
#plt.xlabel('Y coordinate')
#plt.hist(pos_y,bins=20)
#plt.subplot(2,3,3)
#plt.xlabel('photon wavelength')
#plt.hist(ray_l,bins=200)
#plt.subplot(2,3,4)
#plt.xlabel('X direction')
#plt.hist(dir_x,bins=20)
#plt.subplot(2,3,5)
#plt.xlabel('Y direction')
#plt.hist(dir_y,bins=20)
#plt.subplot(2,3,6)
#plt.xlabel('Z direction')
#plt.hist(dir_z,bins=20)
#plt.suptitle(str(photon_num)+' photons')
#plt.show()
