#Piezotronic effect on the luminescence of 
#quantum dots for micro/nano-newton force 
#measurement
#
#Zhang 2017

import numpy as np
import scipy.integrate as integrate
import scipy.constants
import math
import matplotlib.pyplot as plt
import warnings

warnings.simplefilter('ignore', RuntimeWarning)

#################################################
#		Load Constants			#
#################################################
hbar=scipy.constants.hbar			#
pi=scipy.constants.pi				#
e=scipy.constants.e				#
eps_0=scipy.constants.epsilon_0			#
kb=scipy.constants.k				#
c=scipy.constants.c				#
bohr=scipy.constants.value('Bohr radius')	#
electron_mass=scipy.constants.m_e		#
#################################################

#################################################################
#			Input Parameters			#
#################################################################
eps_r=5.8               #relative permittivity                  #
r=2.060e-9/bohr		#QD radius				#
Eg=1.75/27.2		#Bandgap energy in Hartree		#
Eex=1.8/(eps_r*r)	#exciton energy (depends on size of QD)	#
me=0.12			#electron relative mass in au		#
mh=0.25			#hole relative mass in au		#
d=3.92e-12*8.24e-8/e	#piezoelectric constant in bohr/hartree	#
F_min=6.10e-8/8.24e-8	#starting force in N (scaled - hart/b)	#
F_max=7.10e-8/8.24e-8	#end force				#
F_step=0.1e-8/8.24e-8	#force resolution			#
gamma=11.e-3/27.2	#electron-photon coupling in Hartree	#
temp_min=300./315775.	#starting temperature in Hartree	#
temp_max=300./315775.	#end temperature			#
temp_step=1./315775.	#temperature resolution			#
method='linear'		#Varshni or linear temperature		#
			#dependence (see Quantum dots to	#
			#probe temperature and pressure		#
			#in highly confined liquids		#
			#Albahrani 2018 Eq(2)			#
varshni_a=0.		#Varshni alpha term			#
varshni_b=0.		#Varshni beta term			#
linear_a=-0.32*315775/27200.					#
			#linear_a=-0.32meV/K			#
#mu0e=3.0		#initial guess of e chem pot in Hartree	#
#mu0h=2.0		#initial guess of h chem pot		#
#mu_method='fermi'	#fermi or e_gas method to obtain	#
#			#chemical potential in self		#
#			#consistent loop			#
#			#(see Zhang method section)		#
muthresh=1.e16*(bohr**3)#threshold for mu self consistency	#
			#this is in carrier density distance	#
max_it=100		#number of iterations			#
wavelength_min=400.	#minimum wavelength calculated in nm	#
wavelength_max=1200.	#maximum wavelength			#
wavelength_step=0.100	#wavelength resolution			#
set_n=2.68e21*(bohr**3)	#exciton density in bohr^-3		#
mu_search_min=-10.	#lower limit of brute force search in eV#
mu_search_max=10.	#upper limit of brute force search in eV#
mu_search_step=0.01	#brute force mu search resolution	#
#################################################################

#########################################################################
#		Non-Excitonic Quantum Dot Energy Function		#
#########################################################################
def QD_energy(force_i,temp_i,ex):	#ex=y for excitonic, n otherwise#
	Ee=((pi)**2)/(2.*me*r*r)	#e confinement energy (Hartree)	#
	Eh=((pi)**2)/(2.*mh*r*r)	#h confinement energy		#
	print('Electron and Hole confinement energies are:')		#
	print(Ee*27.2,Eh*27.2)						#
	print('eV')							#
	Fe=4.*pi*d*force_i/(eps_r*((2.*r)**2))				#
					#piezo induced displace field	#
					#Stress = force / (2*r)^2	#
	Ep=24.*((2./(3.*pi))**6)*((Fe)**2)*(me+mh)*((2.*r)**4)		#
	                                #piezoelectic energy shift on   #
	                                #electrons and holes            #
	print('Piezoelectric energy is:')				#
	print(Ep*27.2)							#
	print('eV')							#
	if method=='linear':            #check which temp method        #
		Et=linear_a*temp_i	#to decide how to handle temp   #
	else:                           #dependence                     #
		Et=-varshni_a*(temp_i**2)/(temp_i+varshni_b)		#
	print('Total energy of the QD:')				#
	if ex=='y':							#
		EQD=Eg+Ee+Eh-Ep-Eex+Et					#
	else:								#
		EQD=Eg+Ee+Eh-Ep+Et					#
	print(EQD*27.2)							#
	print('eV')							#
	print('Center of Lorentzian at :'+str(45.564/EQD)+' nm')	#
	print(' ')							#
	return EQD,Ep			#Q dot total energy             #
#########################################################################

#################################################################################
#			Chemical Potential Function				#
#################################################################################
def integrand(E_i,temp_i,mu_i):			#first, define density integrad #
						#scale sqrt to Hartree		#
						#everything in exp can stay eV	#
						#obtain mu in eV		#
	return np.sqrt(E_i)/((np.exp((E_i-mu_i)/(temp_i)))+1.)			#
def n_integral(temp_i,mu_i,m_eff):		#function to preform integral	#
	I=integrate.quad(integrand,0.,np.inf,args=(temp_i,mu_i))		#
	return ((2.*m_eff)**1.5)*(1./(2.*pi*pi))*I[0]				#
#################################################################################

#################################################################################
#				Fermi-Dirac Function				#
#################################################################################
def fdf(e_i,mu_i,temp_i):				#define the F-D funciton#
	return 1./(np.exp((e_i-mu_i)/(temp_i))+1.)	#takes energy,mu,temp	#
#################################################################################

#################################################################################################################
#					Brute Force Chemical Potential Search					#
#################################################################################################################
def brute(temp_i,m_i):				#Brute force search for election/hole chemical potential	#
	diff_list=[]				#preallocate list for difference between calc'ed value and 	#
						#desired value							#
	for mu_i in np.arange(mu_search_min,mu_search_max+mu_search_step,mu_search_step):			#
						#for all possible values of chem pot in the desired range	#
		n_new=n_integral(temp_i,mu_i,m_i)								#
						#calculate the integral for carrier density			#
		diff=abs(set_n-n_new)		#check difference between integral density and desired density	#
		diff_list.append(diff)		#add to list							#
	smallest=min(diff_list)			#find the smallest difference in the list			#
	q=diff_list.index(smallest)		#find the index of the smallest difference			#
	print('Found Chemical Potential at')									#
	print(np.arange(mu_search_min,mu_search_max+mu_search_step,mu_search_step)[q])				#
	print('eV')												#
	print('with distance from true carrier density')							#
	print(diff_list[q])											#
	print('bohr^-3')											#
	return np.arange(mu_search_min,mu_search_max+mu_search_step,mu_search_step)[q]				#
#################################################################################################################

#################################################################################
#			Iterative Chemical Potential Search			#
#################################################################################
def chem_pot_search(temp_i,m_i,n_i):						#
	mu_max=((3.*pi*pi*n_i)**(2./3.))/(2.*m_i)				#
	mu_min=temp_i*math.log((n_i/2.)*(2.*pi/(2.*m_i*temp_i)**(1.5)))		#
	#if mu_min<0:								#
	#	mu_min=0.							#
	iteration=max_it							#
	while iteration > 0:							#
		iteration=iteration-1						#
		print('Iteration '+str(max_it-iteration)+\
			': limits are '+str(mu_max*27.2)+','+str(mu_min*27.2))	#
		mu_av=(mu_max+mu_min)/2.					#
		n_new=n_integral(temp_i,mu_av,m_i)				#
		diff=n_new-n_i							#
		print('distance is '+str(diff/(bohr**3))+str(' m^-3'))		#
		if abs(diff)<muthresh:						#
			print('Convergence reached')				#
			break							#
		if diff>0:							#
			mu_max=mu_av						#
		if diff<0:							#
			mu_min=mu_av						#
	print('###############################################################')#
	print('End of search')							#
	print('Chemical potential is estimated to be:')				#
	print(mu_av*27.2)							#
	print('eV')								#
	print('')								#
	return(mu_av)								#
#################################################################################
	

####################################################################################################
####################################################################################################

				#	MAIN CODE STARTS HERE	#

####################################################################################################
####################################################################################################

plt.figure(1)
wavelength=np.arange(wavelength_min,wavelength_max+wavelength_step,wavelength_step)
temp_ex=.1/315775.
temp=temp_min   #just pressure for now
print('###############################################################')
print('# Calculating exciton conduction electron chemical potential  #')
print('###############################################################')
ex_mu_e=chem_pot_search(temp_ex,me,set_n)+Eg
print('###############################################################')
print('#     Calculating exciton valence hole chemical potential     #')
print('###############################################################')
ex_mu_h=chem_pot_search(temp_ex,mh,set_n)
print('###############################################################')
print('Exciton Electron and Hole Chemical Potential are:')
print(ex_mu_e*27.2,ex_mu_h*27.2)
print('eV')
print(' ')
print('###############################################################')
print('#              Constant Temperature Calculations              #')
print('###############################################################')
print(' ')
print('Starting Stress dependent Luminescence Calculations')
e_qd_list=[]
e_p_list=[]
Z=[]
g=open('emi_spec.txt','w+')
for force in np.arange(F_min,F_max,F_step):#+F_step,F_step):
	g.write('Force '+str(8.24e-8*force)+'\n')
	print(' ')
	print('Force = '+str(round(8.24e-8*force,8))+' N')
	nonex_E_calc=QD_energy(force,temp,'n')
	nonex_E_qd=nonex_E_calc[0]
	nonex_E_p=nonex_E_calc[1]
	ex_E_qd=QD_energy(force,temp_ex,'y')[0]
	e_qd_list.append(nonex_E_qd)
	e_p_list.append(nonex_E_p)
	n_induced=3.*d*force/((2.*r)**3)
	print('Ratio of Excitonic to Non-Excitonic Carrier Density is: '+str(set_n/n_induced))
	print('###############################################################')
	print('#Calculating nonexciton conduction electron chemical potential#')
	print('###############################################################')
	nonex_mu_e=chem_pot_search(temp_min,me,n_induced)+Eg
	print('###############################################################')
	print('#   Calculating nonexciton valence hole chemical potential    #')
	print('###############################################################')
	nonex_mu_h=chem_pot_search(temp_min,mh,n_induced)
	print('###############################################################')
	print('Electron and Hole Chemical Potential are:')
	print(nonex_mu_e*27.2,nonex_mu_h*27.2)
	print('eV')
	lum=[]
	for LAMBDA in wavelength:
		value=n_induced*(((gamma/2.)/((((45.56/LAMBDA)-nonex_E_qd)**2)+((gamma/2.)**2)))*fdf(45.56/LAMBDA,nonex_mu_e,temp)*(1.-fdf(45.56/LAMBDA,nonex_mu_h,temp)))+set_n*(((gamma/2.)/((((45.56/LAMBDA)-ex_E_qd)**2)+((gamma/2.)**2))))
		lum.append(value)
#		lum.append(set_n*(((gamma/2.)/((((45.56/LAMBDA)-ex_E_qd)**2)+((gamma/2.)**2)))))
#		lum.append(n_induced*(((gamma/2.)/((((45.56/LAMBDA)-nonex_E_qd)**2)+((gamma/2.)**2))))+set_n*(((gamma/2.)/((((45.56/LAMBDA)-ex_E_qd)**2)+((gamma/2.)**2)))))
		g.write(str(LAMBDA)+' '+str(value)+'\n')
	plt.plot(wavelength,lum)
	Z.append(lum)
plt.xlim(450,600)
plt.show()
X,Y=np.meshgrid(np.arange(F_min*8.24e-2,F_max*8.24e-2+F_step*8.24e-2,F_step*8.24e-2),wavelength)
Z=np.array(Z).T
plt.figure(2)
plt.pcolor(X,Y,Z,shading='auto')
plt.ylabel('Wavelength [nm]')
plt.xlabel('Force [uN]')
plt.show()

plt.figure(3)
plt.plot(np.arange(F_min*8.24e-8,F_max*8.24e-8+F_step*8.24e-8,F_step*8.24e-8),e_qd_list,label='$E_{QD}$')
plt.plot(np.arange(F_min*8.24e-8,F_max*8.24e-8+F_step*8.24e-8,F_step*8.24e-8),e_p_list,label='$E_p$')
plt.xlabel('Force [N]')
plt.ylabel('Energy [eV]')
plt.legend()
plt.show()

